#include <stdatomic.h>

#define MAX_NODES			0xf

typedef unsigned long long pointer;
typedef atomic_ullong pointer_t;

#define MAKE_POINTER(ptr, count)	((((pointer)count) << 32) | ptr)
#define PTR_MASK 0xffffffffLL
#define COUNT_MASK (0xffffffffLL << 32)

static inline void set_count(pointer *p, unsigned int val) { *p = (*p & ~COUNT_MASK) | ((pointer)val << 32); }
static inline void set_ptr(pointer *p, unsigned int val) { *p = (*p & ~PTR_MASK) | val; }
static inline unsigned int get_count(pointer p) { return (p & COUNT_MASK) >> 32; }
static inline unsigned int get_ptr(pointer p) { return p & PTR_MASK; }

typedef struct node {
	unsigned int value;
	pointer_t next;
} node_t;

typedef struct {
	pointer_t head;
	pointer_t tail;
	node_t nodes[MAX_NODES + 1];
} queue_t;

void init_queue(queue_t *q, int num_threads);

/**
	@Begin
	@Global_define:
		@DeclareVar:
		typedef struct tag_elem {
			Tag id;
			unsigned int data;
			
			tag_elem(Tag _id, unsigned int _data) {
				id = _id;
				data = _data;
			}
		} tag_elem_t;

		spec_queue<tag_elem_t> queue;
		Tag tag;
		@InitVar:
			queue = spec_queue<tag_elem_t>();
			tag = Tag();
	@Happens_before:
		# Only check the happens-before relationship according to the id of the
		# commit_point_set. For commit_point_set that has same ID, A -> B means
		# B happens after the previous A.
		Enqueue -> Dequeue
	@End
*/



/**
	@Begin
	@Interface: Enqueue
	@Commit_point_set: Enqueue_Success_Point
	@ID: __sequential.tag.getCurAndInc()
	@Action:
		# __ID__ is an internal macro that refers to the id of the current
		# interface call
		@Code:
		__sequential.queue.enqueue(tag_elem_t(__ID__, val));
	@End
*/
void enqueue(queue_t *q, unsigned int val);

/**
	@Begin
	@Interface: Dequeue
	@Commit_point_set: Dequeue_Success_Point
	@ID: __sequential.queue.peak().tag
	@Action:
		@Code:
		unsigned int _Old_Val = __sequential.queue.dequeue().data;
	@Post_check:
		_Old_Val == __RET__
	@End
*/
unsigned int dequeue(queue_t *q);
int get_thread_num();
